import { existsSync, mkdirSync } from "fs";
import { access, constants, stat } from "fs/promises";
import ms from "ms";
import { sep } from "path";
import { inspect } from "util";
import { testLinking } from "./action.js";
import { validateUArrLs } from "./arr.js";
import { getClient } from "./clients/TorrentClient.js";
import { CrossSeedError } from "./errors.js";
import { Label, logger } from "./logger.js";
import { getRuntimeConfig } from "./runtimeConfig.js";
import { validateTorznabUrls } from "./torznab.js";
import { wait } from "./utils.js";
/**
 * validates existence, permission, and that a path is a directory
 * @param path string of path to validate
 * @param optionName name of the configuration key
 * @param permissions number (see constants in calling function) of permission
 * @returns true if path exists and has required permission
 */
async function verifyPath(path, optionName, permissions) {
    try {
        if ((await stat(path)).isDirectory()) {
            await access(path, permissions);
            return true;
        }
    }
    catch (error) {
        if (error.code === "ENOENT") {
            logger.error(`\tYour ${optionName} "${path}" is not a valid directory on the filesystem.`);
            if (sep === "\\" && !path.includes("\\") && !path.includes("/")) {
                logger.error("\tIt may not be formatted properly for Windows.\n" +
                    '\t\t\t\tMake sure to use "\\\\" or "/" for directory separators.');
            }
        }
        else {
            logger.error(`\tYour ${optionName} "${path}" has invalid permissions.`);
        }
    }
    return false;
}
/**
 * verifies the config paths provided against the filesystem
 * @returns true (if paths are valid)
 */
async function checkConfigPaths() {
    const { dataDirs, injectDir, linkDirs, outputDir, torrentDir } = getRuntimeConfig();
    const READ_ONLY = constants.R_OK;
    const READ_AND_WRITE = constants.R_OK | constants.W_OK;
    let pathFailure = 0;
    if (torrentDir &&
        !(await verifyPath(torrentDir, "torrentDir", READ_ONLY))) {
        pathFailure++;
    }
    if (!(await verifyPath(outputDir, "outputDir", READ_AND_WRITE))) {
        pathFailure++;
    }
    for (const linkDir of linkDirs) {
        if (!existsSync(linkDir)) {
            logger.info(`Creating linkDir: ${linkDir}`);
            mkdirSync(linkDir, { recursive: true });
        }
        if (!(await verifyPath(linkDir, "linkDir", READ_AND_WRITE))) {
            pathFailure++;
        }
    }
    for (const dataDir of dataDirs) {
        if (!(await verifyPath(dataDir, "dataDirs", READ_ONLY))) {
            pathFailure++;
        }
    }
    if (injectDir) {
        logger.warn({
            label: Label.INJECT,
            message: `Manually injecting torrents performs minimal filtering which slightly increases chances of false positives, see the docs for more info`,
        });
        if (!(await verifyPath(injectDir, "injectDir", READ_AND_WRITE))) {
            pathFailure++;
        }
    }
    if (linkDirs.length) {
        for (const dataDir of dataDirs) {
            try {
                testLinking(dataDir);
            }
            catch (e) {
                logger.error(e);
                logger.error("Failed to link from dataDirs to linkDir.");
                pathFailure++;
            }
        }
    }
    if (pathFailure) {
        throw new CrossSeedError(`\tYour configuration is invalid, please see the ${pathFailure > 1 ? "errors" : "error"} above for details.`);
    }
}
async function retry(cb, retries, delayMs) {
    let lastError = new Error("Retry failed");
    for (let i = 0; i <= retries; i++) {
        try {
            return await cb();
        }
        catch (e) {
            const retryMsg = i < retries ? `, retrying in ${delayMs / 1000} seconds` : "";
            logger.error(`Attempt ${i + 1}/${retries + 1} failed${retryMsg}: ${e.message}`);
            logger.debug(e);
            lastError = e;
            if (i >= retries)
                break;
            await wait(delayMs);
        }
    }
    throw lastError;
}
export async function doStartupValidation() {
    await checkConfigPaths(); // ensure paths are valid first
    const validateClientConfig = async () => getClient()?.validateConfig(); // preserve `this` class pointer
    const errors = (await Promise.allSettled([
        retry(validateTorznabUrls, 5, ms("1 minute")),
        retry(validateUArrLs, 5, ms("1 minute")),
        retry(validateClientConfig, 5, ms("1 minute")),
    ])).filter((p) => p.status === "rejected");
    if (errors.length) {
        throw new CrossSeedError(`\tYour configuration is invalid, please see the ${errors.length > 1 ? "errors" : "error"} above for details.`);
    }
    logger.verbose({
        label: Label.CONFIGDUMP,
        message: inspect(getRuntimeConfig()),
    });
    logger.info("Your configuration is valid!");
}
//# sourceMappingURL=startup.js.map